import boto3
import json
import logging
from shelvery.aws_helper import AwsHelper
from datetime import datetime

logger = logging.getLogger(__name__)


class ShelveryQueue:

    def __init__(self, queue_url, wait_period):
        self.queue_url = queue_url
        # Max wait time is 900, if is set to greater, set value to 900
        self.wait_period = int(wait_period) if int(wait_period) < 900 else 900
        logger.info(f"Initialized sqs service with message delay of {self.wait_period} seconds")
        self.sqs = AwsHelper.boto3_client('sqs')

    def send(self, message):
        if isinstance(message, dict):
            message['Timestamp'] = datetime.now().strftime("%Y-%m-%d %H:%M:%S UTC")
            message = json.dumps(message)

        if self.queue_url is not None:
            try:
                response = self.sqs.send_message(
                    QueueUrl=self.queue_url,
                    DelaySeconds=self.wait_period,
                    MessageBody=message
                )
            except:
                logger.exception('Failed to send message to sqs queue')
                logger.error(f"Message:{message}")



import boto3

from typing import List

from botocore.exceptions import ClientError
from shelvery.aws_helper import AwsHelper
from shelvery.engine import SHELVERY_DO_BACKUP_TAGS
from shelvery.ec2_backup import ShelveryEC2Backup
from shelvery.entity_resource import EntityResource
from shelvery.backup_resource import BackupResource


class ShelveryEBSBackup(ShelveryEC2Backup):
    """Shelvery engine implementation for EBS data backups"""

    def __init__(self):
        ShelveryEC2Backup.__init__(self)

    def delete_backup(self, backup_resource: BackupResource):
        ec2client = AwsHelper.boto3_client('ec2', arn=self.role_arn, external_id=self.role_external_id)
        ec2client.delete_snapshot(SnapshotId=backup_resource.backup_id)

    def get_existing_backups(self, tag_prefix: str) -> List[BackupResource]:
        ec2client = AwsHelper.boto3_client('ec2', arn=self.role_arn, external_id=self.role_external_id)
        # lookup snapshots by tags
        snapshots = ec2client.describe_snapshots(Filters=[
            {'Name': f"tag:{tag_prefix}:{BackupResource.BACKUP_MARKER_TAG}", 'Values': ['true']}
        ])
        backups = []

        # create backup resource objects
        for snap in snapshots['Snapshots']:        
            snap_tags = dict(map(lambda t: (t['Key'], t['Value']), snap['Tags']))
            if f"{tag_prefix}:ami_id" in snap_tags:
                self.logger.info(f"EBS snapshot {snap['SnapshotId']} created by AMI shelvery backup, skiping...")
                continue
                            
            backup = BackupResource.construct(
                tag_prefix=tag_prefix,
                backup_id=snap['SnapshotId'],
                tags=snap_tags
            )
            # legacy code - entity id should be picked up from tags
            if backup.entity_id is None:
                self.logger.info(f"SnapshotId is None, using VolumeId {snap['VolumeId']}")
                backup.entity_id = snap['VolumeId']
            backups.append(backup)

        self.populate_volume_information(backups)

        return backups

    def get_engine_type(self) -> str:
        return 'ebs'

    def get_resource_type(self) -> str:
        return 'ec2 volume'

    def backup_resource(self, backup_resource: BackupResource) -> BackupResource:
        ec2client = AwsHelper.boto3_client('ec2', arn=self.role_arn, external_id=self.role_external_id)
        # create snapshot
        snap = ec2client.create_snapshot(
            VolumeId=backup_resource.entity_id,
            Description=backup_resource.name
        )
        backup_resource.backup_id = snap['SnapshotId']
        return backup_resource

    def get_backup_resource(self, region: str, backup_id: str) -> BackupResource:
        ec2 = AwsHelper.boto3_session('ec2', region_name=region, arn=self.role_arn, external_id=self.role_external_id)
        snapshot = ec2.Snapshot(backup_id)
        d_tags = dict(map(lambda t: (t['Key'], t['Value']), snapshot.tags))
        return BackupResource.construct(d_tags['shelvery:tag_name'], backup_id, d_tags)

    def get_entities_to_backup(self, tag_name: str) -> List[EntityResource]:
        volumes = self.collect_volumes(tag_name)
        return list(
            map(
                lambda vol: EntityResource(
                    resource_id=vol['VolumeId'],
                    resource_region=self.region,
                    date_created=vol['CreateTime'],
                    tags=dict(map(lambda t: (t['Key'], t['Value']), vol['Tags']))
                ),
                volumes
            )
        )

    def is_backup_available(self, region: str, backup_id: str) -> bool:
        try:
            regional_client = AwsHelper.boto3_client('ec2', region_name=region, arn=self.role_arn, external_id=self.role_external_id)
            snapshot = regional_client.describe_snapshots(SnapshotIds=[backup_id])['Snapshots'][0]
            complete = snapshot['State'] == 'completed'
            self.logger.info(f"{backup_id} is {snapshot['Progress']} complete")
            return complete
        except Exception as e:
            self.logger.warn(f"Problem getting status of ec2 snapshot status for snapshot {backup_id}:{e}")

    def copy_backup_to_region(self, backup_id: str, region: str):
        ec2client = AwsHelper.boto3_client('ec2', arn=self.role_arn, external_id=self.role_external_id)
        snapshot = ec2client.describe_snapshots(SnapshotIds=[backup_id])['Snapshots'][0]
        regional_client = AwsHelper.boto3_client('ec2', region_name=region, arn=self.role_arn, external_id=self.role_external_id)
        copy_snapshot_response = regional_client.copy_snapshot(SourceSnapshotId=backup_id,
                                                               SourceRegion=ec2client._client_config.region_name,
                                                               DestinationRegion=region,
                                                               Description=snapshot['Description'])

        # return id of newly created snapshot in dr region
        return copy_snapshot_response['SnapshotId']

    def share_backup_with_account(self, backup_region: str, backup_id: str, aws_account_id: str):
        ec2 = AwsHelper.boto3_session('ec2', region_name=backup_region, arn=self.role_arn, external_id=self.role_external_id)
        snapshot = ec2.Snapshot(backup_id)
        snapshot.modify_attribute(Attribute='createVolumePermission',
                                  CreateVolumePermission={
                                      'Add': [{'UserId': aws_account_id}]
                                  },
                                  UserIds=[aws_account_id],
                                  OperationType='add')

    def copy_shared_backup(self, source_account: str, source_backup: BackupResource):
        ec2client = AwsHelper.boto3_client('ec2', arn=self.role_arn, external_id=self.role_external_id)
        snap = ec2client.copy_snapshot(
            SourceSnapshotId=source_backup.backup_id,
            SourceRegion=source_backup.region
        )
        return snap['SnapshotId']
    # collect all volumes tagged with given tag, in paginated manner
    def collect_volumes(self, tag_name: str):
        load_volumes = True
        next_token = ''
        all_volumes = []
        ec2client = AwsHelper.boto3_client('ec2', arn=self.role_arn, external_id=self.role_external_id)
        while load_volumes:
            tagged_volumes = ec2client.describe_volumes(
                Filters=[{'Name': f"tag:{tag_name}", 'Values': SHELVERY_DO_BACKUP_TAGS}],
                NextToken=next_token
            )
            all_volumes = all_volumes + tagged_volumes['Volumes']
            if 'NextToken' in tagged_volumes and len(tagged_volumes['NextToken']) > 0:
                load_volumes = True
                next_token = tagged_volumes['NextToken']
            else:
                load_volumes = False

        return all_volumes

    def populate_volume_information(self, backups):
        volume_ids = []
        volumes = {}
        ec2client = AwsHelper.boto3_client('ec2', arn=self.role_arn, external_id=self.role_external_id)
        local_region = boto3.session.Session().region_name

        # create list of all volume ids
        for backup in backups:
            if backup.entity_id not in volume_ids:
                volume_ids.append(backup.entity_id)

        # populate map volumeid->volume if present
        for volume_id in volume_ids:
            try:
                volume = ec2client.describe_volumes(VolumeIds=[volume_id])['Volumes'][0]
                d_tags = dict(map(lambda t: (t['Key'], t['Value']), volume['Tags']))
                volumes[volume_id] = EntityResource(volume_id, local_region, volume['CreateTime'], d_tags)
            except ClientError as e:
                if 'InvalidVolume.NotFound' in str(e):
                    volumes[volume_id] = EntityResource.empty()
                    volumes[volume_id].resource_id = volume_id
                else:
                    raise e

        # add info to backup resource objects
        for backup in backups:
            if backup.entity_id in volumes:
                backup.entity_resource = volumes[backup.entity_id]



import boto3
import os
import json
import logging

from typing import Dict
from threading import Thread

from shelvery.runtime_config import RuntimeConfig
from shelvery.aws_helper import AwsHelper
from shelvery.queue import ShelveryQueue

class ShelveryInvoker:
    """Helper to orchestrate execution of shelvery operations on AWS Lambda platform"""

    def invoke_shelvery_operation(self, engine, method_name: str, method_arguments: Dict):
        """
        Invokes shelvery engine asynchronously
        If shelvery is running within lambda environment, new lambda function invocation will be made. If running
        on server, it will start new thread and invoke the function
        Function invoke must accept arguments in form of map
        """
        is_lambda_context = RuntimeConfig.is_lambda_runtime(engine)
        is_offload_queueing = RuntimeConfig.is_offload_queueing(engine)
        parameters = {
            'backup_type': engine.get_engine_type(),
            'action': method_name,
            'arguments': method_arguments
        }
        if is_lambda_context:
            if 'config' in engine.lambda_payload:
                parameters['config'] = engine.lambda_payload['config']

            if is_offload_queueing:
                sqs = ShelveryQueue(RuntimeConfig.get_sqs_queue_url(engine),RuntimeConfig.get_sqs_queue_wait_period(engine))
                sqs.send(parameters)
            else:
                parameters['is_started_internally'] = True
                payload = json.dumps(parameters)
                bytes_payload = bytearray()
                bytes_payload.extend(map(ord, payload))
                function_name = os.environ['AWS_LAMBDA_FUNCTION_NAME']
                lambda_client = AwsHelper.boto3_client('lambda')
                lambda_client.invoke_async(FunctionName=function_name, InvokeArgs=bytes_payload)
        else:
            resource_type = engine.get_engine_type()

            def execute():
                from shelvery.factory import ShelveryFactory
                backup_engine = ShelveryFactory.get_shelvery_instance(resource_type)
                method = backup_engine.__getattribute__(method_name)
                method(method_arguments)

            logging.info(f"Start new thread to execute :{method_name}")
            if 'SHELVERY_MONO_THREAD' in os.environ and os.environ['SHELVERY_MONO_THREAD'] == "1":
                execute()
            else:
                thread = Thread(target=execute)
                thread.start()
            # thread.join()



import logging
from shelvery.factory import ShelveryFactory


class ShelveryCliMain:
    
    def main(self, backup_type, action):
    
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        
        # create backup engine
        backup_engine = ShelveryFactory.get_shelvery_instance(backup_type)
        method = backup_engine.__getattribute__(action)
        
        # start the action
        method()
        return 0




from datetime import datetime
from typing import Dict
import boto3


class EntityResource:
    """Represents entity such as ec2 volume, instance or rds instance"""
    
    def __init__(self, resource_id: str, resource_region: str, date_created: datetime, tags: Dict):
        self.resource_id = resource_id
        self.date_created = date_created
        self.tags = tags
        self.resource_region = resource_region
    
    @classmethod
    def empty(cls):
        local_region = boto3.session.Session().region_name
        resource = EntityResource(None, local_region, None, {})
        return resource
